@@ -12,8 +12,10 @@
 /// assert_eq!(lower_n_mask(2), 0b11);
 /// # }
 /// ```
-#[inline]
-pub fn lower_n_mask(n: u64) -> u64 {
+#[must_use]
+#[inline(always)]
+#[allow(clippy::match_bool)] // reason="easier to visualize logic"
+pub const fn lower_n_mask(n: u64) -> u64 {
     debug_assert!(n <= 64, "lower_n_mask() overflow in shl.");
 
     match n == 64 {
@@ -32,8 +34,10 @@ pub fn lower_n_mask(n: u64) -> u64 {
 /// assert_eq!(lower_n_halfway(2), 0b10);
 /// # }
 /// ```
-#[inline]
-pub fn lower_n_halfway(n: u64) -> u64 {
+#[must_use]
+#[inline(always)]
+#[allow(clippy::match_bool)] // reason="easier to visualize logic"
+pub const fn lower_n_halfway(n: u64) -> u64 {
     debug_assert!(n <= 64, "lower_n_halfway() overflow in shl.");
 
     match n == 0 {
@@ -52,8 +56,9 @@ pub fn lower_n_halfway(n: u64) -> u64 {
 /// assert_eq!(nth_bit(2), 0b100);
 /// # }
 /// ```
-#[inline]
-pub fn nth_bit(n: u64) -> u64 {
+#[must_use]
+#[inline(always)]
+pub const fn nth_bit(n: u64) -> u64 {
     debug_assert!(n < 64, "nth_bit() overflow in shl.");
     1 << n
 }
