@@ -3,8 +3,25 @@
 //! This is adapted from [fast-float-rust](https://github.com/aldanor/fast-float-rust),
 //! a port of [fast_float](https://github.com/fastfloat/fast_float) to Rust.
 
+// NOTE: We never want to disable multi-digit optimizations when parsing our floats,
+// since the nanoseconds it saves on branching is irrelevant when considering decimal
+// points and fractional digits and it majorly improves longer floats.
+
 #![doc(hidden)]
 
+#[cfg(not(feature = "compact"))]
+use lexical_parse_integer::algorithm;
+#[cfg(feature = "f16")]
+use lexical_util::bf16::bf16;
+use lexical_util::digit::{char_to_digit_const, char_to_valid_digit_const};
+use lexical_util::error::Error;
+#[cfg(feature = "f16")]
+use lexical_util::f16::f16;
+use lexical_util::format::NumberFormat;
+use lexical_util::iterator::{AsBytes, Bytes, DigitsIter, Iter};
+use lexical_util::result::Result;
+use lexical_util::step::u64_step;
+
 #[cfg(any(feature = "compact", feature = "radix"))]
 use crate::bellerophon::bellerophon;
 #[cfg(feature = "power-of-two")]
@@ -16,18 +33,6 @@ use crate::number::Number;
 use crate::options::Options;
 use crate::shared;
 use crate::slow::slow_radix;
-#[cfg(not(feature = "compact"))]
-use lexical_parse_integer::algorithm;
-#[cfg(feature = "f16")]
-use lexical_util::bf16::bf16;
-use lexical_util::digit::{char_to_digit_const, char_to_valid_digit_const};
-use lexical_util::error::Error;
-#[cfg(feature = "f16")]
-use lexical_util::f16::f16;
-use lexical_util::format::NumberFormat;
-use lexical_util::iterator::{AsBytes, Bytes, BytesIter};
-use lexical_util::result::Result;
-use lexical_util::step::u64_step;
 
 // API
 // ---
@@ -82,14 +87,16 @@ pub trait ParseFloat: LemireFloat {
         parse_partial::<Self, FORMAT>(bytes, options)
     }
 
-    /// Forward complete parser parameters to the backend, using only the fast path.
+    /// Forward complete parser parameters to the backend, using only the fast
+    /// path.
     #[cfg_attr(not(feature = "compact"), inline(always))]
     fn fast_path_complete<const FORMAT: u128>(bytes: &[u8], options: &Options) -> Result<Self> {
         check_radix!(FORMAT);
         fast_path_complete::<Self, FORMAT>(bytes, options)
     }
 
-    /// Forward partial parser parameters to the backend, using only the fast path.
+    /// Forward partial parser parameters to the backend, using only the fast
+    /// path.
     #[cfg_attr(not(feature = "compact"), inline(always))]
     fn fast_path_partial<const FORMAT: u128>(
         bytes: &[u8],
@@ -154,42 +161,44 @@ parse_float_as_f32! { bf16 f16 }
 // NOTE:
 //  The partial and complete parsers are done separately because it provides
 //  minor optimizations when parsing invalid input, and the logic is slightly
-//  different internally. Most of the code is reshared, so the duplicated
+//  different internally. Most of the code is shared, so the duplicated
 //  code is only like 30 lines.
 
-macro_rules! parse_mantissa_sign {
-    ($byte:ident, $format:ident) => {{
-        match $byte.integer_iter().peek() {
-            Some(&b'+') if !$format.no_positive_mantissa_sign() => (false, 1),
-            Some(&b'+') if $format.no_positive_mantissa_sign() => {
-                return Err(Error::InvalidPositiveSign($byte.cursor()));
-            },
-            Some(&b'-') => (true, 1),
-            Some(_) if $format.required_mantissa_sign() => {
-                return Err(Error::MissingSign($byte.cursor()));
-            },
-            _ => (false, 0),
-        }
-    }};
+/// Parse the sign from the leading digits.
+#[cfg_attr(not(feature = "compact"), inline(always))]
+pub fn parse_mantissa_sign<const FORMAT: u128>(byte: &mut Bytes<'_, FORMAT>) -> Result<bool> {
+    let format = NumberFormat::<{ FORMAT }> {};
+    parse_sign!(
+        byte,
+        true,
+        format.no_positive_mantissa_sign(),
+        format.required_mantissa_sign(),
+        InvalidPositiveSign,
+        MissingSign
+    )
 }
 
-macro_rules! parse_exponent_sign {
-    ($byte:ident, $format:ident) => {{
-        match $byte.integer_iter().peek() {
-            Some(&b'+') if !$format.no_positive_exponent_sign() => (false, 1),
-            Some(&b'+') if $format.no_positive_exponent_sign() => {
-                return Err(Error::InvalidPositiveExponentSign($byte.cursor()));
-            },
-            Some(&b'-') => (true, 1),
-            Some(_) if $format.required_mantissa_sign() => {
-                return Err(Error::MissingExponentSign($byte.cursor()));
-            },
-            _ => (false, 0),
-        }
-    }};
+/// Parse the sign from the leading digits.
+#[cfg_attr(not(feature = "compact"), inline(always))]
+pub fn parse_exponent_sign<const FORMAT: u128>(byte: &mut Bytes<'_, FORMAT>) -> Result<bool> {
+    let format = NumberFormat::<{ FORMAT }> {};
+    parse_sign!(
+        byte,
+        true,
+        format.no_positive_exponent_sign(),
+        format.required_exponent_sign(),
+        InvalidPositiveExponentSign,
+        MissingExponentSign
+    )
 }
 
 /// Utility to extract the result and handle any errors from parsing a `Number`.
+///
+/// - `format` - The numerical format as a packed integer
+/// - `byte` - The `DigitsIter` iterator
+/// - `is_negative` - If the final value is negative
+/// - `parse_normal` - The function to parse non-special numbers with
+/// - `parse_special` - The function to parse special numbers with
 macro_rules! parse_number {
     (
         $format:ident,
@@ -215,6 +224,9 @@ macro_rules! parse_number {
 }
 
 /// Convert extended float to native.
+///
+/// - `type` - The native floating point type.
+/// - `fp` - The extended floating-point representation.
 macro_rules! to_native {
     ($type:ident, $fp:ident, $is_negative:ident) => {{
         let mut float = extended_to_float::<$type>($fp);
@@ -226,21 +238,26 @@ macro_rules! to_native {
 }
 
 /// Parse a float from bytes using a complete parser.
+#[allow(clippy::missing_inline_in_public_items)] // reason = "only public for testing"
 pub fn parse_complete<F: LemireFloat, const FORMAT: u128>(
     bytes: &[u8],
     options: &Options,
 ) -> Result<F> {
-    let format = NumberFormat::<{ FORMAT }> {};
     let mut byte = bytes.bytes::<{ FORMAT }>();
-    let (is_negative, shift) = parse_mantissa_sign!(byte, format);
-    // SAFETY: safe since we shift at most one for a parsed sign byte.
-    unsafe { byte.step_by_unchecked(shift) };
+    let is_negative = parse_mantissa_sign(&mut byte)?;
     if byte.integer_iter().is_consumed() {
-        return Err(Error::Empty(byte.cursor()));
+        if NumberFormat::<FORMAT>::REQUIRED_INTEGER_DIGITS
+            || NumberFormat::<FORMAT>::REQUIRED_MANTISSA_DIGITS
+        {
+            return Err(Error::Empty(byte.cursor()));
+        } else {
+            return Ok(F::ZERO);
+        }
     }
 
     // Parse our a small representation of our number.
-    let num = parse_number!(FORMAT, byte, is_negative, options, parse_number, parse_special);
+    let num: Number<'_> =
+        parse_number!(FORMAT, byte, is_negative, options, parse_complete_number, parse_special);
     // Try the fast-path algorithm.
     if let Some(value) = num.try_fast_path::<_, FORMAT>() {
         return Ok(value);
@@ -252,7 +269,7 @@ pub fn parse_complete<F: LemireFloat, const FORMAT: u128>(
     // Fallback to a slower, but always correct algorithm. If we have
     // lossy, we can't be here.
     if fp.exp < 0 {
-        debug_assert!(!options.lossy());
+        debug_assert!(!options.lossy(), "lossy algorithms never use slow algorithms");
         // Undo the invalid extended float biasing.
         fp.exp -= shared::INVALID_FP;
         fp = slow_path::<F, FORMAT>(num, fp);
@@ -263,36 +280,45 @@ pub fn parse_complete<F: LemireFloat, const FORMAT: u128>(
 }
 
 /// Parse a float using only the fast path as a complete parser.
+#[allow(clippy::missing_inline_in_public_items)] // reason = "only public for testing"
 pub fn fast_path_complete<F: LemireFloat, const FORMAT: u128>(
     bytes: &[u8],
     options: &Options,
 ) -> Result<F> {
-    let format = NumberFormat::<{ FORMAT }> {};
     let mut byte = bytes.bytes::<{ FORMAT }>();
-    let (is_negative, shift) = parse_mantissa_sign!(byte, format);
-    // SAFETY: safe since we shift at most one for a parsed sign byte.
-    unsafe { byte.step_by_unchecked(shift) };
+    let is_negative = parse_mantissa_sign(&mut byte)?;
     if byte.integer_iter().is_consumed() {
-        return Err(Error::Empty(byte.cursor()));
+        if NumberFormat::<FORMAT>::REQUIRED_INTEGER_DIGITS
+            || NumberFormat::<FORMAT>::REQUIRED_MANTISSA_DIGITS
+        {
+            return Err(Error::Empty(byte.cursor()));
+        } else {
+            return Ok(F::ZERO);
+        }
     }
 
     // Parse our a small representation of our number.
-    let num = parse_number!(FORMAT, byte, is_negative, options, parse_number, parse_special);
+    let num =
+        parse_number!(FORMAT, byte, is_negative, options, parse_complete_number, parse_special);
     Ok(num.force_fast_path::<_, FORMAT>())
 }
 
 /// Parse a float from bytes using a partial parser.
+#[allow(clippy::missing_inline_in_public_items)] // reason = "only public for testing"
 pub fn parse_partial<F: LemireFloat, const FORMAT: u128>(
     bytes: &[u8],
     options: &Options,
 ) -> Result<(F, usize)> {
-    let format = NumberFormat::<{ FORMAT }> {};
     let mut byte = bytes.bytes::<{ FORMAT }>();
-    let (is_negative, shift) = parse_mantissa_sign!(byte, format);
-    // SAFETY: safe since we shift at most one for a parsed sign byte.
-    unsafe { byte.step_by_unchecked(shift) };
+    let is_negative = parse_mantissa_sign(&mut byte)?;
     if byte.integer_iter().is_consumed() {
-        return Err(Error::Empty(byte.cursor()));
+        if NumberFormat::<FORMAT>::REQUIRED_INTEGER_DIGITS
+            || NumberFormat::<FORMAT>::REQUIRED_MANTISSA_DIGITS
+        {
+            return Err(Error::Empty(byte.cursor()));
+        } else {
+            return Ok((F::ZERO, byte.cursor()));
+        }
     }
 
     // Parse our a small representation of our number.
@@ -315,7 +341,7 @@ pub fn parse_partial<F: LemireFloat, const FORMAT: u128>(
     // Fallback to a slower, but always correct algorithm. If we have
     // lossy, we can't be here.
     if fp.exp < 0 {
-        debug_assert!(!options.lossy());
+        debug_assert!(!options.lossy(), "lossy algorithms never use slow algorithms");
         // Undo the invalid extended float biasing.
         fp.exp -= shared::INVALID_FP;
         fp = slow_path::<F, FORMAT>(num, fp);
@@ -326,17 +352,21 @@ pub fn parse_partial<F: LemireFloat, const FORMAT: u128>(
 }
 
 /// Parse a float using only the fast path as a partial parser.
+#[allow(clippy::missing_inline_in_public_items)] // reason = "only public for testing"
 pub fn fast_path_partial<F: LemireFloat, const FORMAT: u128>(
     bytes: &[u8],
     options: &Options,
 ) -> Result<(F, usize)> {
-    let format = NumberFormat::<{ FORMAT }> {};
     let mut byte = bytes.bytes::<{ FORMAT }>();
-    let (is_negative, shift) = parse_mantissa_sign!(byte, format);
-    // SAFETY: safe since we shift at most one for a parsed sign byte.
-    unsafe { byte.step_by_unchecked(shift) };
+    let is_negative = parse_mantissa_sign(&mut byte)?;
     if byte.integer_iter().is_consumed() {
-        return Err(Error::Empty(byte.cursor()));
+        if NumberFormat::<FORMAT>::REQUIRED_INTEGER_DIGITS
+            || NumberFormat::<FORMAT>::REQUIRED_MANTISSA_DIGITS
+        {
+            return Err(Error::Empty(byte.cursor()));
+        } else {
+            return Ok((F::ZERO, byte.cursor()));
+        }
     }
 
     // Parse our a small representation of our number.
@@ -356,7 +386,8 @@ pub fn fast_path_partial<F: LemireFloat, const FORMAT: u128>(
 
 /// Wrapper for different moderate-path algorithms.
 /// A return exponent of `-1` indicates an invalid value.
-#[inline]
+#[must_use]
+#[inline(always)]
 pub fn moderate_path<F: LemireFloat, const FORMAT: u128>(
     num: &Number,
     lossy: bool,
@@ -418,7 +449,8 @@ pub fn moderate_path<F: LemireFloat, const FORMAT: u128>(
 
 /// Invoke the slow path.
 /// At this point, the float string has already been validated.
-#[inline]
+#[must_use]
+#[inline(always)]
 pub fn slow_path<F: LemireFloat, const FORMAT: u128>(
     num: Number,
     fp: ExtendedFloat80,
@@ -446,9 +478,13 @@ pub fn slow_path<F: LemireFloat, const FORMAT: u128>(
 ///
 /// This creates a representation of the float as the
 /// significant digits and the decimal exponent.
-#[inline]
-#[allow(clippy::collapsible_if)]
-pub fn parse_partial_number<'a, const FORMAT: u128>(
+#[cfg_attr(not(feature = "compact"), inline(always))]
+#[allow(unused_mut)] // reason = "used when format is enabled"
+#[allow(clippy::unwrap_used)] // reason = "developer error if we incorrectly assume an overflow"
+#[allow(clippy::collapsible_if)] // reason = "more readable uncollapsed"
+#[allow(clippy::cast_possible_wrap)] // reason = "no hardware supports buffers >= i64::MAX"
+#[allow(clippy::too_many_lines)] // reason = "function is one logical entity"
+pub fn parse_number<'a, const FORMAT: u128, const IS_PARTIAL: bool>(
     mut byte: Bytes<'a, FORMAT>,
     is_negative: bool,
     options: &Options,
@@ -461,7 +497,8 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
     //      We've tried:
     //          - checking for explicit overflow, via `overflowing_mul`.
     //          - counting the max number of steps.
-    //          - subslicing the string, and only processing the first `step` digits.
+    //          - subslicing the string, and only processing the first `step`
+    //            digits.
     //          - pre-computing the maximum power, and only adding until then.
     //
     //      All of these lead to substantial performance penalty.
@@ -483,34 +520,31 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
     let format = NumberFormat::<{ FORMAT }> {};
     let decimal_point = options.decimal_point();
     let exponent_character = options.exponent();
-    debug_assert!(format.is_valid());
-    debug_assert!(!byte.is_done());
+    debug_assert!(format.is_valid(), "should have already checked for an invalid number format");
+    debug_assert!(!byte.is_buffer_empty(), "should have previously checked for empty input");
     let bits_per_digit = shared::log2(format.mantissa_radix()) as i64;
     let bits_per_base = shared::log2(format.exponent_base()) as i64;
 
     // INTEGER
 
     // Check to see if we have a valid base prefix.
-    let base_prefix = format.base_prefix();
+    #[allow(unused_variables)]
     let mut is_prefix = false;
-    let mut iter = byte.integer_iter();
-    if cfg!(feature = "format") && base_prefix != 0 && iter.peek() == Some(&b'0') {
-        // SAFETY: safe since `byte.len() >= 1`.
-        unsafe { iter.step_unchecked() };
-        // Check to see if the next character is the base prefix.
-        // We must have a format like `0x`, `0d`, `0o`. Note:
-        if let Some(&c) = iter.peek() {
-            is_prefix = if format.case_sensitive_base_prefix() {
-                c == base_prefix
-            } else {
-                c.to_ascii_lowercase() == base_prefix.to_ascii_lowercase()
-            };
-            if is_prefix {
-                // SAFETY: safe since `byte.len() >= 1`.
-                unsafe { iter.step_unchecked() };
-                if iter.is_done() {
-                    return Err(Error::Empty(iter.cursor()));
-                }
+    #[cfg(feature = "format")]
+    {
+        let base_prefix = format.base_prefix();
+        let mut iter = byte.integer_iter();
+        if base_prefix != 0 && iter.read_if_value_cased(b'0').is_some() {
+            // Check to see if the next character is the base prefix.
+            // We must have a format like `0x`, `0d`, `0o`.
+            // NOTE: The check for empty integer digits happens below so
+            // we don't need a redundant check here.
+            is_prefix = true;
+            if iter.read_if_value(base_prefix, format.case_sensitive_base_prefix()).is_some()
+                && iter.is_buffer_empty()
+                && format.required_integer_digits()
+            {
+                return Err(Error::EmptyInteger(iter.cursor()));
             }
         }
     }
@@ -521,21 +555,43 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
     #[cfg(not(feature = "compact"))]
     parse_8digits::<_, FORMAT>(byte.integer_iter(), &mut mantissa);
     parse_digits::<_, _, FORMAT>(byte.integer_iter(), |digit| {
-        mantissa = mantissa.wrapping_mul(format.radix() as _).wrapping_add(digit as _);
+        mantissa = mantissa.wrapping_mul(format.radix() as u64).wrapping_add(digit as u64);
     });
     let mut n_digits = byte.current_count() - start.current_count();
-    if cfg!(feature = "format") && format.required_integer_digits() && n_digits == 0 {
+    #[cfg(feature = "format")]
+    if format.required_integer_digits() && n_digits == 0 {
         return Err(Error::EmptyInteger(byte.cursor()));
     }
 
     // Store the integer digits for slow-path algorithms.
+    // NOTE: We can't use the number of digits to extract the slice for
+    // non-contiguous iterators, but we also need to the number of digits
+    // for our value calculation. We store both, and let the compiler know
+    // to optimize it out when not needed.
+    let b_digits = if cfg!(feature = "format") && !byte.integer_iter().is_contiguous() {
+        byte.cursor() - start.cursor()
+    } else {
+        n_digits
+    };
+    debug_assert!(
+        b_digits <= start.as_slice().len(),
+        "number of digits parsed must <= buffer length"
+    );
     // SAFETY: safe, since `n_digits <= start.as_slice().len()`.
-    debug_assert!(n_digits <= start.as_slice().len());
-    let integer_digits = unsafe { start.as_slice().get_unchecked(..n_digits) };
+    // This is since `byte.len() >= start.len()` but has to have
+    // the same end bounds (that is, `start = byte.clone()`), so
+    // `0 <= byte.current_count() <= start.current_count() <= start.lent()`
+    // so, this will always return only the integer digits.
+    //
+    // NOTE: Removing this code leads to ~10% reduction in parsing
+    // that triggers the Eisell-Lemire algorithm or the digit comp
+    // algorithms, so don't remove the unsafe indexing.
+    let integer_digits = unsafe { start.as_slice().get_unchecked(..b_digits) };
 
     // Check if integer leading zeros are disabled.
-    if cfg!(feature = "format") && !is_prefix && format.no_float_leading_zeros() {
-        if integer_digits.len() > 1 && integer_digits.get(0) == Some(&b'0') {
+    #[cfg(feature = "format")]
+    if !is_prefix && format.no_float_leading_zeros() {
+        if integer_digits.len() > 1 && integer_digits.first() == Some(&b'0') {
             return Err(Error::InvalidLeadingZeros(start.cursor()));
         }
     }
@@ -548,70 +604,94 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
     let mut implicit_exponent: i64;
     let int_end = n_digits as i64;
     let mut fraction_digits = None;
-    if byte.first_is(decimal_point) {
-        // SAFETY: s cannot be empty due to first_is
+    let has_decimal = byte.first_is_cased(decimal_point);
+    if has_decimal {
+        // SAFETY: byte cannot be empty due to `first_is`
         unsafe { byte.step_unchecked() };
         let before = byte.clone();
         #[cfg(not(feature = "compact"))]
         parse_8digits::<_, FORMAT>(byte.fraction_iter(), &mut mantissa);
         parse_digits::<_, _, FORMAT>(byte.fraction_iter(), |digit| {
-            mantissa = mantissa.wrapping_mul(format.radix() as _).wrapping_add(digit as _);
+            mantissa = mantissa.wrapping_mul(format.radix() as u64).wrapping_add(digit as u64);
         });
         n_after_dot = byte.current_count() - before.current_count();
+        // NOTE: We can't use the number of digits to extract the slice for
+        // non-contiguous iterators, but we also need to the number of digits
+        // for our value calculation. We store both, and let the compiler know
+        // to optimize it out when not needed.
+        let b_after_dot = if cfg!(feature = "format") && !byte.fraction_iter().is_contiguous() {
+            byte.cursor() - before.cursor()
+        } else {
+            n_after_dot
+        };
 
         // Store the fraction digits for slow-path algorithms.
-        // SAFETY: safe, since `n_after_dot <= before.as_slice().len()`.
-        debug_assert!(n_after_dot <= before.as_slice().len());
-        fraction_digits = Some(unsafe { before.as_slice().get_unchecked(..n_after_dot) });
+        debug_assert!(
+            b_after_dot <= before.as_slice().len(),
+            "digits after dot must be smaller than buffer"
+        );
+        // SAFETY: safe, since `idx_after_dot <= before.as_slice().len()`.
+        fraction_digits = Some(unsafe { before.as_slice().get_unchecked(..b_after_dot) });
 
         // Calculate the implicit exponent: the number of digits after the dot.
         implicit_exponent = -(n_after_dot as i64);
         if format.mantissa_radix() == format.exponent_base() {
             exponent = implicit_exponent;
         } else {
-            debug_assert!(bits_per_digit % bits_per_base == 0);
+            debug_assert!(bits_per_digit % bits_per_base == 0, "exponent must be a power of base");
             exponent = implicit_exponent * bits_per_digit / bits_per_base;
         };
-        if cfg!(feature = "format") && format.required_fraction_digits() && n_after_dot == 0 {
+        #[cfg(feature = "format")]
+        if format.required_fraction_digits() && n_after_dot == 0 {
             return Err(Error::EmptyFraction(byte.cursor()));
         }
     }
 
+    // NOTE: Check if we have our exponent **BEFORE** checking if the
+    // mantissa is empty, so we can ensure
+    let has_exponent = byte
+        .first_is(exponent_character, format.case_sensitive_exponent() && cfg!(feature = "format"));
+
+    // check to see if we have any invalid leading zeros
     n_digits += n_after_dot;
-    if format.required_mantissa_digits() && n_digits == 0 {
-        return Err(Error::EmptyMantissa(byte.cursor()));
+    if format.required_mantissa_digits()
+        && (n_digits == 0 || (cfg!(feature = "format") && byte.current_count() == 0))
+    {
+        let any_digits = start.clone().integer_iter().peek().is_some();
+        // NOTE: This is because numbers like `_12.34` have significant digits,
+        // they just don't have a valid digit (#97).
+        if has_decimal || has_exponent || !any_digits || IS_PARTIAL {
+            return Err(Error::EmptyMantissa(byte.cursor()));
+        } else {
+            return Err(Error::InvalidDigit(start.cursor()));
+        }
     }
 
     // EXPONENT
 
     // Handle scientific notation.
     let mut explicit_exponent = 0_i64;
-    let is_exponent = if cfg!(feature = "format") && format.case_sensitive_exponent() {
-        byte.first_is(exponent_character)
-    } else {
-        byte.case_insensitive_first_is(exponent_character)
-    };
-    if is_exponent {
+    if has_exponent {
+        // NOTE: See above for the safety invariant above `required_mantissa_digits`.
+        // This is separated for correctness concerns, and therefore the two cannot
+        // be on the same line.
+        // SAFETY: byte cannot be empty due to `first_is` from `has_exponent`.`
+        unsafe { byte.step_unchecked() };
+
         // Check float format syntax checks.
-        if cfg!(feature = "format") {
+        #[cfg(feature = "format")]
+        {
+            // NOTE: We've overstepped for the safety invariant before.
             if format.no_exponent_notation() {
-                return Err(Error::InvalidExponent(byte.cursor()));
+                return Err(Error::InvalidExponent(byte.cursor() - 1));
             }
             // Check if we have no fraction but we required exponent notation.
             if format.no_exponent_without_fraction() && fraction_digits.is_none() {
-                return Err(Error::ExponentWithoutFraction(byte.cursor()));
+                return Err(Error::ExponentWithoutFraction(byte.cursor() - 1));
             }
         }
 
-        // SAFETY: byte cannot be empty due to first_is
-        unsafe { byte.step_unchecked() };
-        let (is_negative, shift) = parse_exponent_sign!(byte, format);
-        // SAFETY: safe since we shift at most one for a parsed sign byte.
-        unsafe { byte.step_by_unchecked(shift) };
-        if cfg!(feature = "format") && format.required_exponent_sign() && shift == 0 {
-            return Err(Error::MissingExponentSign(byte.cursor()));
-        }
-
+        let is_negative_exponent = parse_exponent_sign(&mut byte)?;
         let before = byte.current_count();
         parse_digits::<_, _, FORMAT>(byte.exponent_iter(), |digit| {
             if explicit_exponent < 0x10000000 {
@@ -623,7 +703,7 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
             return Err(Error::EmptyExponent(byte.cursor()));
         }
         // Handle our sign, and get the explicit part of the exponent.
-        explicit_exponent = if is_negative {
+        explicit_exponent = if is_negative_exponent {
             -explicit_exponent
         } else {
             explicit_exponent
@@ -636,14 +716,11 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
     // Check to see if we have a valid base suffix.
     // We've already trimmed any leading digit separators here, so we can be safe
     // that the first character **is not** a digit separator.
+    #[allow(unused_variables)]
     let base_suffix = format.base_suffix();
-    if cfg!(feature = "format") && base_suffix != 0 {
-        let is_suffix = if cfg!(feature = "format") && format.case_sensitive_base_suffix() {
-            byte.first_is(base_suffix)
-        } else {
-            byte.case_insensitive_first_is(base_suffix)
-        };
-        if is_suffix {
+    #[cfg(feature = "format")]
+    if base_suffix != 0 {
+        if byte.first_is(base_suffix, format.case_sensitive_base_suffix()) {
             // SAFETY: safe since `byte.len() >= 1`.
             unsafe { byte.step_unchecked() };
         }
@@ -655,7 +732,8 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
     let end = byte.cursor();
     let mut step = u64_step(format.radix());
     let mut many_digits = false;
-    if cfg!(feature = "format") && !format.required_mantissa_digits() && n_digits == 0 {
+    #[cfg(feature = "format")]
+    if !format.required_mantissa_digits() && n_digits == 0 {
         exponent = 0;
     }
     if n_digits <= step {
@@ -676,21 +754,13 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
     n_digits -= step;
     let mut zeros = start.clone();
     let mut zeros_integer = zeros.integer_iter();
-    while zeros_integer.peek_is(b'0') {
-        n_digits = n_digits.saturating_sub(1);
-        // SAFETY: safe since zeros cannot be empty due to peek_is
-        unsafe { zeros_integer.step_unchecked() };
-    }
-    if zeros.first_is(decimal_point) {
-        // SAFETY: safe since zeros cannot be empty due to first_is
+    n_digits = n_digits.saturating_sub(zeros_integer.skip_zeros());
+    if zeros.first_is_cased(decimal_point) {
+        // SAFETY: safe since zeros cannot be empty due to `first_is`
         unsafe { zeros.step_unchecked() };
     }
     let mut zeros_fraction = zeros.fraction_iter();
-    while zeros_fraction.peek_is(b'0') {
-        n_digits = n_digits.saturating_sub(1);
-        // SAFETY: safe since zeros cannot be empty due to peek_is
-        unsafe { zeros_fraction.step_unchecked() };
-    }
+    n_digits = n_digits.saturating_sub(zeros_fraction.skip_zeros());
 
     // OVERFLOW
 
@@ -704,7 +774,17 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
         let mut integer_iter = integer.integer_iter();
         integer_iter.skip_zeros();
         parse_u64_digits::<_, FORMAT>(integer_iter, &mut mantissa, &mut step);
-        implicit_exponent = if step == 0 {
+        // NOTE: With the format feature enabled and non-contiguous iterators, we can
+        // have null fraction digits even if step was not 0. We want to make the
+        // none check as late in there as possible: any of them should
+        // short-circuit and should be determined at compile time. So, the
+        // conditions are either:
+        // 1. Step == 0
+        // 2. `cfg!(feature = "format") && !byte.is_contiguous() &&
+        //    fraction_digits.is_none()`
+        implicit_exponent = if step == 0
+            || (cfg!(feature = "format") && !byte.is_contiguous() && fraction_digits.is_none())
+        {
             // Filled our mantissa with just the integer.
             int_end - integer.current_count() as i64
         } else {
@@ -725,7 +805,7 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
         if format.mantissa_radix() == format.exponent_base() {
             exponent = implicit_exponent;
         } else {
-            debug_assert!(bits_per_digit % bits_per_base == 0);
+            debug_assert!(bits_per_digit % bits_per_base == 0, "exponent must be a power of base");
             exponent = implicit_exponent * bits_per_digit / bits_per_base;
         };
         // Add back the explicit exponent.
@@ -745,15 +825,24 @@ pub fn parse_partial_number<'a, const FORMAT: u128>(
     ))
 }
 
+pub fn parse_partial_number<'a, const FORMAT: u128>(
+    byte: Bytes<'a, FORMAT>,
+    is_negative: bool,
+    options: &Options,
+) -> Result<(Number<'a>, usize)> {
+    parse_number::<FORMAT, true>(byte, is_negative, options)
+}
+
 /// Try to parse a non-special floating point number.
-#[inline]
-pub fn parse_number<'a, const FORMAT: u128>(
+#[inline(always)]
+pub fn parse_complete_number<'a, const FORMAT: u128>(
     byte: Bytes<'a, FORMAT>,
     is_negative: bool,
     options: &Options,
 ) -> Result<Number<'a>> {
-    let length = byte.length();
-    let (float, count) = parse_partial_number::<FORMAT>(byte, is_negative, options)?;
+    // Then have a const `IsPartial` as well
+    let length = byte.buffer_length();
+    let (float, count) = parse_number::<FORMAT, false>(byte, is_negative, options)?;
     if count == length {
         Ok(float)
     } else {
@@ -765,10 +854,10 @@ pub fn parse_number<'a, const FORMAT: u128>(
 // ------
 
 /// Iteratively parse and consume digits from bytes.
-#[inline]
+#[inline(always)]
 pub fn parse_digits<'a, Iter, Cb, const FORMAT: u128>(mut iter: Iter, mut cb: Cb)
 where
-    Iter: BytesIter<'a>,
+    Iter: DigitsIter<'a>,
     Cb: FnMut(u32),
 {
     let format = NumberFormat::<{ FORMAT }> {};
@@ -779,23 +868,25 @@ where
             None => break,
         }
         // SAFETY: iter cannot be empty due to `iter.peek()`.
+        // NOTE: Because of the match statement, this would optimize poorly with
+        // `read_if`.
         unsafe { iter.step_unchecked() };
+        iter.increment_count();
     }
 }
 
 /// Iteratively parse and consume digits in intervals of 8.
-#[inline]
+#[inline(always)]
 #[cfg(not(feature = "compact"))]
 pub fn parse_8digits<'a, Iter, const FORMAT: u128>(mut iter: Iter, mantissa: &mut u64)
 where
-    Iter: BytesIter<'a>,
+    Iter: DigitsIter<'a>,
 {
     let format = NumberFormat::<{ FORMAT }> {};
     let radix: u64 = format.radix() as u64;
-    if can_try_parse_8digits!(iter, radix) {
-        let radix2 = radix.wrapping_mul(radix);
-        let radix4 = radix2.wrapping_mul(radix2);
-        let radix8 = radix4.wrapping_mul(radix4);
+    if can_try_parse_multidigit!(iter, radix) {
+        debug_assert!(radix < 16, "radices over 16 will overflow with radix^8");
+        let radix8 = format.radix8() as u64;
         // Can do up to 2 iterations without overflowing, however, for large
         // inputs, this is much faster than any other alternative.
         while let Some(v) = algorithm::try_parse_8digits::<u64, _, FORMAT>(&mut iter) {
@@ -809,23 +900,22 @@ where
 /// # Preconditions
 ///
 /// There must be at least `step` digits left in iterator.
-#[inline]
+#[cfg_attr(not(feature = "compact"), inline(always))]
 pub fn parse_u64_digits<'a, Iter, const FORMAT: u128>(
     mut iter: Iter,
     mantissa: &mut u64,
     step: &mut usize,
 ) where
-    Iter: BytesIter<'a>,
+    Iter: DigitsIter<'a>,
 {
     let format = NumberFormat::<{ FORMAT }> {};
     let radix = format.radix() as u64;
 
     // Try to parse 8 digits at a time, if we can.
     #[cfg(not(feature = "compact"))]
-    if can_try_parse_8digits!(iter, radix) {
-        let radix2 = radix.wrapping_mul(radix);
-        let radix4 = radix2.wrapping_mul(radix2);
-        let radix8 = radix4.wrapping_mul(radix4);
+    if can_try_parse_multidigit!(iter, radix) {
+        debug_assert!(radix < 16, "radices over 16 will overflow with radix^8");
+        let radix8 = format.radix8() as u64;
         while *step > 8 {
             if let Some(v) = algorithm::try_parse_8digits::<u64, _, FORMAT>(&mut iter) {
                 *mantissa = mantissa.wrapping_mul(radix8).wrapping_add(v);
@@ -844,6 +934,7 @@ pub fn parse_u64_digits<'a, Iter, const FORMAT: u128>(
             *step -= 1;
             // SAFETY: safe, since `iter` cannot be empty due to `iter.peek()`.
             unsafe { iter.step_unchecked() };
+            iter.increment_count();
         } else {
             break;
         }
@@ -855,7 +946,8 @@ pub fn parse_u64_digits<'a, Iter, const FORMAT: u128>(
 
 /// Determine if the input data matches the special string.
 /// If there's no match, returns 0. Otherwise, returns the byte's cursor.
-#[inline]
+#[must_use]
+#[inline(always)]
 pub fn is_special_eq<const FORMAT: u128>(mut byte: Bytes<FORMAT>, string: &'static [u8]) -> usize {
     let format = NumberFormat::<{ FORMAT }> {};
     if cfg!(feature = "format") && format.case_sensitive_special() {
@@ -864,7 +956,7 @@ pub fn is_special_eq<const FORMAT: u128>(mut byte: Bytes<FORMAT>, string: &'stat
             byte.special_iter().peek();
             return byte.cursor();
         }
-    } else if shared::case_insensitive_starts_with(byte.special_iter(), string.iter()) {
+    } else if shared::starts_with_uncased(byte.special_iter(), string.iter()) {
         // Trim the iterator afterwards.
         byte.special_iter().peek();
         return byte.cursor();
@@ -873,7 +965,8 @@ pub fn is_special_eq<const FORMAT: u128>(mut byte: Bytes<FORMAT>, string: &'stat
 }
 
 /// Parse a positive representation of a special, non-finite float.
-#[inline]
+#[must_use]
+#[cfg_attr(not(feature = "compact"), inline(always))]
 pub fn parse_positive_special<F, const FORMAT: u128>(
     byte: Bytes<FORMAT>,
     options: &Options,
@@ -887,7 +980,7 @@ where
     }
 
     let cursor = byte.cursor();
-    let length = byte.length() - cursor;
+    let length = byte.buffer_length() - cursor;
     if let Some(nan_string) = options.nan_string() {
         if length >= nan_string.len() {
             let count = is_special_eq::<FORMAT>(byte.clone(), nan_string);
@@ -917,7 +1010,8 @@ where
 }
 
 /// Parse a partial representation of a special, non-finite float.
-#[inline]
+#[must_use]
+#[inline(always)]
 pub fn parse_partial_special<F, const FORMAT: u128>(
     byte: Bytes<FORMAT>,
     is_negative: bool,
@@ -934,7 +1028,8 @@ where
 }
 
 /// Try to parse a special, non-finite float.
-#[inline]
+#[must_use]
+#[inline(always)]
 pub fn parse_special<F, const FORMAT: u128>(
     byte: Bytes<FORMAT>,
     is_negative: bool,
@@ -943,7 +1038,7 @@ pub fn parse_special<F, const FORMAT: u128>(
 where
     F: LemireFloat,
 {
-    let length = byte.length();
+    let length = byte.buffer_length();
     if let Some((float, count)) = parse_partial_special::<F, FORMAT>(byte, is_negative, options) {
         if count == length {
             return Some(float);
