@@ -6,23 +6,25 @@
 
 #![doc(hidden)]
 
-#[cfg(feature = "radix")]
-use crate::bigint::Bigfloat;
-use crate::bigint::{Bigint, Limb, LIMB_BITS};
-use crate::float::{extended_to_float, ExtendedFloat80, RawFloat};
-use crate::limits::{u32_power_limit, u64_power_limit};
-use crate::number::Number;
-use crate::shared;
 use core::cmp;
+
 #[cfg(not(feature = "compact"))]
 use lexical_parse_integer::algorithm;
 use lexical_util::digit::char_to_valid_digit_const;
 #[cfg(feature = "radix")]
 use lexical_util::digit::digit_to_char_const;
 use lexical_util::format::NumberFormat;
-use lexical_util::iterator::{AsBytes, BytesIter};
+use lexical_util::iterator::{AsBytes, DigitsIter, Iter};
 use lexical_util::num::{AsPrimitive, Integer};
 
+#[cfg(feature = "radix")]
+use crate::bigint::Bigfloat;
+use crate::bigint::{Bigint, Limb};
+use crate::float::{extended_to_float, ExtendedFloat80, RawFloat};
+use crate::limits::{u32_power_limit, u64_power_limit};
+use crate::number::Number;
+use crate::shared;
+
 // ALGORITHM
 // ---------
 
@@ -42,14 +44,16 @@ use lexical_util::num::{AsPrimitive, Integer};
 /// any value before or equal to `16777217.0` must be rounded down
 /// to `16777216.0`. These near-halfway conversions therefore may require
 /// a large number of digits to unambiguously determine how to round.
-#[inline]
+#[must_use]
+#[inline(always)]
+#[allow(clippy::unwrap_used)] // reason = "none is a developer error"
 pub fn slow_radix<F: RawFloat, const FORMAT: u128>(
     num: Number,
     fp: ExtendedFloat80,
 ) -> ExtendedFloat80 {
     // Ensure our preconditions are valid:
     //  1. The significant digits are not shifted into place.
-    debug_assert!(fp.mant & (1 << 63) != 0);
+    debug_assert!(fp.mant & (1 << 63) != 0, "number must be normalized");
 
     let format = NumberFormat::<{ FORMAT }> {};
 
@@ -93,7 +97,9 @@ pub fn slow_radix<F: RawFloat, const FORMAT: u128>(
 /// exponent relative to the significant digits, we scale the real
 /// digits to the theoretical digits for `b` and determine if we
 /// need to round-up.
-#[inline]
+#[must_use]
+#[inline(always)]
+#[allow(clippy::cast_possible_wrap)] // reason = "the value range is [-324, 308]"
 pub fn digit_comp<F: RawFloat, const FORMAT: u128>(
     num: Number,
     fp: ExtendedFloat80,
@@ -110,7 +116,12 @@ pub fn digit_comp<F: RawFloat, const FORMAT: u128>(
     }
 }
 
-/// Generate the significant digits with a positive exponent relative to mantissa.
+/// Generate the significant digits with a positive exponent relative to
+/// mantissa.
+#[must_use]
+#[allow(clippy::unwrap_used)] // reason = "none is a developer error"
+#[allow(clippy::cast_possible_wrap)] // reason = "can't wrap in practice: max is ~1000 limbs"
+#[allow(clippy::missing_inline_in_public_items)] // reason = "only public for testing"
 pub fn positive_digit_comp<F: RawFloat, const FORMAT: u128>(
     mut bigmant: Bigint,
     exponent: i32,
@@ -142,7 +153,8 @@ pub fn positive_digit_comp<F: RawFloat, const FORMAT: u128>(
     fp
 }
 
-/// Generate the significant digits with a negative exponent relative to mantissa.
+/// Generate the significant digits with a negative exponent relative to
+/// mantissa.
 ///
 /// This algorithm is quite simple: we have the significant digits `m1 * b^N1`,
 /// where `m1` is the bigint mantissa, `b` is the radix, and `N1` is the radix
@@ -162,7 +174,10 @@ pub fn positive_digit_comp<F: RawFloat, const FORMAT: u128>(
 ///
 /// This allows us to compare both floats using integers efficiently
 /// without any loss of precision.
-#[allow(clippy::comparison_chain)]
+#[allow(clippy::match_bool)] // reason = "simplifies documentation"
+#[allow(clippy::unwrap_used)] // reason = "unwrap panics if a developer error"
+#[allow(clippy::comparison_chain)] // reason = "logically different conditions for algorithm"
+#[allow(clippy::missing_inline_in_public_items)] // reason = "only exposed for unittesting"
 pub fn negative_digit_comp<F: RawFloat, const FORMAT: u128>(
     bigmant: Bigint,
     mut fp: ExtendedFloat80,
@@ -170,7 +185,7 @@ pub fn negative_digit_comp<F: RawFloat, const FORMAT: u128>(
 ) -> ExtendedFloat80 {
     // Ensure our preconditions are valid:
     //  1. The significant digits are not shifted into place.
-    debug_assert!(fp.mant & (1 << 63) != 0);
+    debug_assert!(fp.mant & (1 << 63) != 0, "the significant digits must be normalized");
 
     let format = NumberFormat::<FORMAT> {};
     let radix = format.radix();
@@ -178,7 +193,7 @@ pub fn negative_digit_comp<F: RawFloat, const FORMAT: u128>(
     // Get the significant digits and radix exponent for the real digits.
     let mut real_digits = bigmant;
     let real_exp = exponent;
-    debug_assert!(real_exp < 0);
+    debug_assert!(real_exp < 0, "algorithm only works with negative numbers");
 
     // Round down our extended-precision float and calculate `b`.
     let mut b = fp;
@@ -241,6 +256,15 @@ pub fn negative_digit_comp<F: RawFloat, const FORMAT: u128>(
 }
 
 /// Try to parse 8 digits at a time.
+///
+/// - `format` - The numerical format specification as a packed 128-bit integer
+/// - `iter` - An iterator over all bytes in the buffer
+/// - `value` - The currently parsed value.
+/// - `count` - The total number of parsed digits
+/// - `counter` - The number of parsed digits since creating the current u32
+/// - `step` - The maximum number of digits for the radix that can fit in a u32.
+/// - `max_digits` - The maximum number of digits that can affect floating-point
+///   rounding.
 #[cfg(not(feature = "compact"))]
 macro_rules! try_parse_8digits {
     (
@@ -256,11 +280,9 @@ macro_rules! try_parse_8digits {
         let radix = format.radix() as Limb;
 
         // Try 8-digit optimizations.
-        if can_try_parse_8digits!($iter, radix) {
-            let radix2 = radix.wrapping_mul(radix);
-            let radix4 = radix2.wrapping_mul(radix2);
-            let radix8 = radix4.wrapping_mul(radix4);
-
+        if can_try_parse_multidigit!($iter, radix) {
+            debug_assert!(radix < 16);
+            let radix8 = format.radix8() as Limb;
             while $step - $counter >= 8 && $max_digits - $count >= 8 {
                 if let Some(v) = algorithm::try_parse_8digits::<Limb, _, FORMAT>(&mut $iter) {
                     $value = $value.wrapping_mul(radix8).wrapping_add(v);
@@ -275,6 +297,11 @@ macro_rules! try_parse_8digits {
 }
 
 /// Add a digit to the temporary value.
+///
+/// - `c` - The character to convert to a digit.
+/// - `value` - The currently parsed value.
+/// - `count` - The total number of parsed digits
+/// - `counter` - The number of parsed digits since creating the current u32
 macro_rules! add_digit {
     ($c:ident, $radix:ident, $value:ident, $counter:ident, $count:ident) => {{
         let digit = char_to_valid_digit_const($c, $radix);
@@ -288,6 +315,12 @@ macro_rules! add_digit {
 }
 
 /// Add a temporary value to our mantissa.
+///
+/// - `format` - The numerical format specification as a packed 128-bit integer
+/// - `result` - The big integer,
+/// - `power` - The power to scale the big integer by.
+/// - `value` - The value to add to the big integer,
+/// - `counter` - The number of parsed digits since creating the current u32
 macro_rules! add_temporary {
     // Multiply by the small power and add the native value.
     (@mul $result:ident, $power:expr, $value:expr) => {
@@ -296,14 +329,9 @@ macro_rules! add_temporary {
     };
 
     // Add a temporary where we won't read the counter results internally.
-    //
-    // # Safety
-    //
-    // Safe is `counter <= step`, or smaller than the table size.
     (@end $format:ident, $result:ident, $counter:ident, $value:ident) => {
         if $counter != 0 {
-            // SAFETY: safe, since `counter <= step`, or smaller than the table size.
-            let small_power = unsafe { f64::int_pow_fast_path($counter, $format.radix()) };
+            let small_power = f64::int_pow_fast_path($counter, $format.radix());
             add_temporary!(@mul $result, small_power as Limb, $value);
         }
     };
@@ -317,6 +345,10 @@ macro_rules! add_temporary {
 }
 
 /// Round-up a truncated value.
+///
+/// - `format` - The numerical format specification as a packed 128-bit integer
+/// - `result` - The big integer,
+/// - `count` - The total number of parsed digits
 macro_rules! round_up_truncated {
     ($format:ident, $result:ident, $count:ident) => {{
         // Need to round-up.
@@ -328,14 +360,19 @@ macro_rules! round_up_truncated {
 }
 
 /// Check and round-up the fraction if any non-zero digits exist.
+///
+/// - `format` - The numerical format specification as a packed 128-bit integer
+/// - `iter` - An iterator over all bytes in the buffer
+/// - `result` - The big integer,
+/// - `count` - The total number of parsed digits
 macro_rules! round_up_nonzero {
     ($format:ident, $iter:expr, $result:ident, $count:ident) => {{
-        // NOTE: All digits must be valid.
+        // NOTE: All digits must already be valid.
         let mut iter = $iter;
 
         // First try reading 8-digits at a time.
         if iter.is_contiguous() {
-            while let Some(value) = iter.read::<u64>() {
+            while let Some(value) = iter.peek_u64() {
                 // SAFETY: safe since we have at least 8 bytes in the buffer.
                 unsafe { iter.step_by_unchecked(8) };
                 if value != 0x3030_3030_3030_3030 {
@@ -359,6 +396,9 @@ macro_rules! round_up_nonzero {
 ///
 /// Returns the parsed mantissa and the number of digits in the mantissa.
 /// The max digits is the maximum number of digits plus one.
+#[must_use]
+#[allow(clippy::cognitive_complexity)] // reason = "complexity broken into macros"
+#[allow(clippy::missing_inline_in_public_items)] // reason = "only public for testing"
 pub fn parse_mantissa<const FORMAT: u128>(num: Number, max_digits: usize) -> (Bigint, usize) {
     let format = NumberFormat::<FORMAT> {};
     let radix = format.radix();
@@ -373,7 +413,7 @@ pub fn parse_mantissa<const FORMAT: u128>(num: Number, max_digits: usize) -> (Bi
     let mut result = Bigint::new();
 
     // Now use our pre-computed small powers iteratively.
-    let step = if LIMB_BITS == 32 {
+    let step = if Limb::BITS == 32 {
         u32_power_limit(format.radix())
     } else {
         u64_power_limit(format.radix())
@@ -405,7 +445,7 @@ pub fn parse_mantissa<const FORMAT: u128>(num: Number, max_digits: usize) -> (Bi
             round_up_nonzero!(format, integer_iter, result, count);
             if let Some(fraction) = num.fraction {
                 let mut fraction = fraction.bytes::<FORMAT>();
-                round_up_nonzero!(format, fraction.fraction_iter(), result, count)
+                round_up_nonzero!(format, fraction.fraction_iter(), result, count);
             }
             return (result, count);
         } else {
@@ -459,6 +499,11 @@ pub fn parse_mantissa<const FORMAT: u128>(num: Number, max_digits: usize) -> (Bi
 }
 
 /// Compare actual integer digits to the theoretical digits.
+///
+/// - `iter` - An iterator over all bytes in the buffer
+/// - `num` - The actual digits of the real floating point number.
+/// - `den` - The theoretical digits created by `b+h` to determine if `b` or
+///   `b+1`
 #[cfg(feature = "radix")]
 macro_rules! integer_compare {
     ($iter:ident, $num:ident, $den:ident, $radix:ident) => {{
@@ -492,6 +537,11 @@ macro_rules! integer_compare {
 }
 
 /// Compare actual fraction digits to the theoretical digits.
+///
+/// - `iter` - An iterator over all bytes in the buffer
+/// - `num` - The actual digits of the real floating point number.
+/// - `den` - The theoretical digits created by `b+h` to determine if `b` or
+///   `b+1`
 #[cfg(feature = "radix")]
 macro_rules! fraction_compare {
     ($iter:ident, $num:ident, $den:ident, $radix:ident) => {{
@@ -537,7 +587,8 @@ macro_rules! fraction_compare {
 /// Adapted from "Bigcomp: Deciding Truncated, Near Halfway Conversions",
 /// available [here](https://www.exploringbinary.com/bigcomp-deciding-truncated-near-halfway-conversions/).
 #[cfg(feature = "radix")]
-#[allow(clippy::comparison_chain)]
+#[allow(clippy::unwrap_used)] // reason = "none is a developer error due to shl overflow"
+#[allow(clippy::comparison_chain)] // reason = "logically different conditions for algorithm"
 pub fn byte_comp<F: RawFloat, const FORMAT: u128>(
     number: Number,
     mut fp: ExtendedFloat80,
@@ -564,8 +615,8 @@ pub fn byte_comp<F: RawFloat, const FORMAT: u128>(
     let mut den: Bigfloat;
 
     if sci_exp < 0 {
-        // Need to have the basen factor be the numerator, and the fp
-        // be the denominator. Since we assumed that theor was the numerator,
+        // Need to have the basen factor be the numerator, and the `fp`
+        // be the denominator. Since we assumed that `theor` was the numerator,
         // if it's the denominator, we need to multiply it into the numerator.
         num = factor;
         num.data *= &theor.data;
@@ -594,11 +645,11 @@ pub fn byte_comp<F: RawFloat, const FORMAT: u128>(
         num.shl(shift).unwrap();
         num.exp -= shift as i32;
     } else if diff > 0 {
-        // Need to shift denominator left, go by a power of LIMB_BITS.
+        // Need to shift denominator left, go by a power of Limb::BITS.
         // After this, the numerator will be non-normalized, and the
         // denominator will be normalized. We need to add one to the
         // quotient,since we're calculating the ceiling of the divmod.
-        let (q, r) = shift.ceil_divmod(LIMB_BITS);
+        let (q, r) = shift.ceil_divmod(Limb::BITS as usize);
         let r = -r;
         if r != 0 {
             num.shl_bits(r as usize).unwrap();
@@ -606,7 +657,7 @@ pub fn byte_comp<F: RawFloat, const FORMAT: u128>(
         }
         if q != 0 {
             den.shl_limbs(q).unwrap();
-            den.exp -= LIMB_BITS as i32 * q as i32;
+            den.exp -= Limb::BITS as i32 * q as i32;
         }
     }
 
@@ -628,7 +679,14 @@ pub fn byte_comp<F: RawFloat, const FORMAT: u128>(
 }
 
 /// Compare digits between the generated values the ratio and the actual view.
+///
+/// - `number` - The representation of the float as a big number, with the
+///   parsed digits.
+/// - `num` - The actual digits of the real floating point number.
+/// - `den` - The theoretical digits created by `b+h` to determine if `b` or
+///   `b+1`
 #[cfg(feature = "radix")]
+#[allow(clippy::unwrap_used)] // reason = "none is a developer error due to a missing fraction"
 pub fn compare_bytes<const FORMAT: u128>(
     number: Number,
     mut num: Bigfloat,
@@ -642,7 +700,7 @@ pub fn compare_bytes<const FORMAT: u128>(
     let mut integer = number.integer.bytes::<{ FORMAT }>();
     let mut integer_iter = integer.integer_iter();
     integer_iter.skip_zeros();
-    if integer_iter.is_done() {
+    if integer_iter.is_buffer_empty() {
         // Cannot be empty, since we must have at least **some** significant digits.
         let mut fraction = number.fraction.unwrap().bytes::<{ FORMAT }>();
         let mut fraction_iter = fraction.fraction_iter();
@@ -669,7 +727,8 @@ pub fn compare_bytes<const FORMAT: u128>(
 
 /// Calculate the scientific exponent from a `Number` value.
 /// Any other attempts would require slowdowns for faster algorithms.
-#[inline]
+#[must_use]
+#[inline(always)]
 pub fn scientific_exponent<const FORMAT: u128>(num: &Number) -> i32 {
     // This has the significant digits and exponent relative to those
     // digits: therefore, we just need to scale to mantissa to `[1, radix)`.
@@ -677,7 +736,7 @@ pub fn scientific_exponent<const FORMAT: u128>(num: &Number) -> i32 {
     let format = NumberFormat::<FORMAT> {};
 
     // Use power reduction to make this faster: we need at least
-    // F::MANTISSA_SIZE bits, so we must have at least radix^4 digits.
+    // `F::MANTISSA_SIZE` bits, so we must have at least radix^4 digits.
     // IF we're using base 3, we can have at most 11 divisions, and
     // base 36, at most ~4. So, this is reasonably efficient.
     let radix = format.radix() as u64;
@@ -701,7 +760,8 @@ pub fn scientific_exponent<const FORMAT: u128>(num: &Number) -> i32 {
 }
 
 /// Calculate `b` from a a representation of `b` as a float.
-#[inline]
+#[must_use]
+#[inline(always)]
 pub fn b<F: RawFloat>(float: F) -> ExtendedFloat80 {
     ExtendedFloat80 {
         mant: float.mantissa().as_u64(),
@@ -710,7 +770,8 @@ pub fn b<F: RawFloat>(float: F) -> ExtendedFloat80 {
 }
 
 /// Calculate `b+h` from a a representation of `b` as a float.
-#[inline]
+#[must_use]
+#[inline(always)]
 pub fn bh<F: RawFloat>(float: F) -> ExtendedFloat80 {
     let fp = b(float);
     ExtendedFloat80 {
@@ -719,40 +780,56 @@ pub fn bh<F: RawFloat>(float: F) -> ExtendedFloat80 {
     }
 }
 
+// NOTE: There will never be binary factors here.
+
 /// Calculate the integral ceiling of the binary factor from a basen number.
-#[inline]
+#[must_use]
+#[inline(always)]
+#[cfg(feature = "radix")]
+pub const fn integral_binary_factor(radix: u32) -> u32 {
+    match radix {
+        3 => 2,
+        5 => 3,
+        6 => 3,
+        7 => 3,
+        9 => 4,
+        10 => 4,
+        11 => 4,
+        12 => 4,
+        13 => 4,
+        14 => 4,
+        15 => 4,
+        17 => 5,
+        18 => 5,
+        19 => 5,
+        20 => 5,
+        21 => 5,
+        22 => 5,
+        23 => 5,
+        24 => 5,
+        25 => 5,
+        26 => 5,
+        27 => 5,
+        28 => 5,
+        29 => 5,
+        30 => 5,
+        31 => 5,
+        33 => 6,
+        34 => 6,
+        35 => 6,
+        36 => 6,
+        // Invalid radix
+        _ => 0,
+    }
+}
+
+/// Calculate the integral ceiling of the binary factor from a basen number.
+#[must_use]
+#[inline(always)]
+#[cfg(not(feature = "radix"))]
 pub const fn integral_binary_factor(radix: u32) -> u32 {
     match radix {
-        3 if cfg!(feature = "radix") => 2,
-        5 if cfg!(feature = "radix") => 3,
-        6 if cfg!(feature = "radix") => 3,
-        7 if cfg!(feature = "radix") => 3,
-        9 if cfg!(feature = "radix") => 4,
         10 => 4,
-        11 if cfg!(feature = "radix") => 4,
-        12 if cfg!(feature = "radix") => 4,
-        13 if cfg!(feature = "radix") => 4,
-        14 if cfg!(feature = "radix") => 4,
-        15 if cfg!(feature = "radix") => 4,
-        17 if cfg!(feature = "radix") => 5,
-        18 if cfg!(feature = "radix") => 5,
-        19 if cfg!(feature = "radix") => 5,
-        20 if cfg!(feature = "radix") => 5,
-        21 if cfg!(feature = "radix") => 5,
-        22 if cfg!(feature = "radix") => 5,
-        23 if cfg!(feature = "radix") => 5,
-        24 if cfg!(feature = "radix") => 5,
-        25 if cfg!(feature = "radix") => 5,
-        26 if cfg!(feature = "radix") => 5,
-        27 if cfg!(feature = "radix") => 5,
-        28 if cfg!(feature = "radix") => 5,
-        29 if cfg!(feature = "radix") => 5,
-        30 if cfg!(feature = "radix") => 5,
-        31 if cfg!(feature = "radix") => 5,
-        33 if cfg!(feature = "radix") => 6,
-        34 if cfg!(feature = "radix") => 6,
-        35 if cfg!(feature = "radix") => 6,
-        36 if cfg!(feature = "radix") => 6,
         // Invalid radix
         _ => 0,
     }
